# Copyright (C) 2013 Simon Feltman <sfeltman@gnome.org>
#
#   docstring.py: documentation string generator for gi.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

from ._gi import (
    VFuncInfo,
    FunctionInfo,
    CallableInfo,
    ObjectInfo,
    StructInfo,
    Direction,
    TypeTag,
)


#: Module storage for currently registered doc string generator function.
_generate_doc_string_func = None


def set_doc_string_generator(func):
    """Set doc string generator function.

    :param callable func:
        Callable which takes a GIInfoStruct and returns documentation for it.
    """
    global _generate_doc_string_func
    _generate_doc_string_func = func


def get_doc_string_generator():
    """Returns the currently registered doc string generator."""
    return _generate_doc_string_func


def generate_doc_string(info):
    """Generate a doc string given a GIInfoStruct.

    :param gi.types.BaseInfo info:
        GI info instance to generate documentation for.
    :returns:
        Generated documentation as a string.
    :rtype: str

    This passes the info struct to the currently registered doc string
    generator and returns the result.
    """
    return _generate_doc_string_func(info)


_type_tag_to_py_type = {
    TypeTag.BOOLEAN: bool,
    TypeTag.INT8: int,
    TypeTag.UINT8: int,
    TypeTag.INT16: int,
    TypeTag.UINT16: int,
    TypeTag.INT32: int,
    TypeTag.UINT32: int,
    TypeTag.INT64: int,
    TypeTag.UINT64: int,
    TypeTag.FLOAT: float,
    TypeTag.DOUBLE: float,
    TypeTag.GLIST: list,
    TypeTag.GSLIST: list,
    TypeTag.ARRAY: list,
    TypeTag.GHASH: dict,
    TypeTag.UTF8: str,
    TypeTag.FILENAME: str,
    TypeTag.UNICHAR: str,
    TypeTag.INTERFACE: None,
    TypeTag.GTYPE: None,
    TypeTag.ERROR: None,
    TypeTag.VOID: None,
}


def _get_pytype_hint(gi_type):
    type_tag = gi_type.get_tag()
    py_type = _type_tag_to_py_type.get(type_tag)

    if py_type and hasattr(py_type, "__name__"):
        return py_type.__name__
    if type_tag == TypeTag.INTERFACE:
        iface = gi_type.get_interface()

        info_name = iface.get_name()
        if not info_name:
            return gi_type.get_tag_as_string()

        return f"{iface.get_namespace()}.{info_name}"

    return gi_type.get_tag_as_string()


def _generate_callable_info_doc(info):
    in_args_strs = []
    if isinstance(info, VFuncInfo) or (
        isinstance(info, FunctionInfo) and info.is_method()
    ):
        in_args_strs = ["self"]

    args = info.get_arguments()
    hint_blacklist = ("void",)

    # Build lists of indices prior to adding the docs because it is possible
    # the index retrieved comes before input arguments being used.
    ignore_indices = {info.get_return_type().get_array_length_index()}
    user_data_indices = set()
    for arg in args:
        ignore_indices.add(arg.get_destroy_index())
        ignore_indices.add(arg.get_type_info().get_array_length_index())
        user_data_indices.add(arg.get_closure_index())

    # Build input argument strings
    for i, arg in enumerate(args):
        if arg.get_direction() == Direction.OUT:
            continue  # skip exclusively output args
        if i in ignore_indices:
            continue
        argstr = arg.get_name()
        hint = _get_pytype_hint(arg.get_type_info())
        if hint not in hint_blacklist:
            argstr += ":" + hint
        if arg.may_be_null() or i in user_data_indices:
            # allow-none or user_data from a closure
            argstr += "=None"
        elif arg.is_optional():
            argstr += "=<optional>"
        in_args_strs.append(argstr)
    in_args_str = ", ".join(in_args_strs)

    # Build return + output argument strings
    out_args_strs = []
    return_hint = _get_pytype_hint(info.get_return_type())
    if not info.skip_return() and return_hint and return_hint not in hint_blacklist:
        argstr = return_hint
        if info.may_return_null():
            argstr += " or None"
        out_args_strs.append(argstr)

    for i, arg in enumerate(args):
        if arg.get_direction() == Direction.IN:
            continue  # skip exclusively input args
        if i in ignore_indices:
            continue
        argstr = arg.get_name()
        hint = _get_pytype_hint(arg.get_type_info())
        if hint not in hint_blacklist:
            argstr += ":" + hint
        out_args_strs.append(argstr)

    if out_args_strs:
        return f"{info.__name__}({in_args_str}) -> {', '.join(out_args_strs)}"
    return f"{info.__name__}({in_args_str})"


def _generate_class_info_doc(info):
    header = "\n:Constructors:\n\n::\n\n"  # start with \n to avoid auto indent of other lines
    doc = ""

    if isinstance(info, StructInfo):
        # Don't show default constructor for disguised (0 length) structs
        if info.get_size() > 0:
            doc += "    " + info.get_name() + "()\n"
    else:
        doc += "    " + info.get_name() + "(**properties)\n"

    for method_info in info.get_methods():
        if method_info.is_constructor():
            doc += "    " + _generate_callable_info_doc(method_info) + "\n"

    if doc:
        return header + doc
    return ""


def _generate_doc_dispatch(info):
    if isinstance(info, (ObjectInfo, StructInfo)):
        return _generate_class_info_doc(info)

    if isinstance(info, CallableInfo):
        return _generate_callable_info_doc(info)

    return ""


set_doc_string_generator(_generate_doc_dispatch)



# Copyright (C) 2010 Simon van der Linden <svdlinden@src.gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

from ..overrides import override
from ..module import get_introspection_module

GIMarshallingTests = get_introspection_module("GIMarshallingTests")

__all__ = []

OVERRIDES_CONSTANT = 7
__all__.append("OVERRIDES_CONSTANT")


class OverridesStruct(GIMarshallingTests.OverridesStruct):
    def __new__(cls, long_):
        return GIMarshallingTests.OverridesStruct.__new__(cls)

    def __init__(self, long_):
        GIMarshallingTests.OverridesStruct.__init__(self)
        self.long_ = long_

    def method(self):
        return GIMarshallingTests.OverridesStruct.method(self) / 7


OverridesStruct = override(OverridesStruct)
__all__.append("OverridesStruct")


class OverridesObject(GIMarshallingTests.OverridesObject):
    def __new__(cls, long_):
        return GIMarshallingTests.OverridesObject.__new__(cls)

    def __init__(self, long_):
        GIMarshallingTests.OverridesObject.__init__(self)
        # FIXME: doesn't work yet
        # self.long_ = long_

    @classmethod
    def new(cls, long_):
        # FIXME: doesn't work yet
        # self.long_ = long_
        # return self
        return GIMarshallingTests.OverridesObject.new()

    def method(self):
        """Overridden doc string."""
        return GIMarshallingTests.OverridesObject.method(self) / 7


OverridesObject = override(OverridesObject)
__all__.append("OverridesObject")



# Copyright (C) 2024 James Henstridge <james@jamesh.id.au>
#
#   _signature.py: callable signature generator for gi.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

from importlib import import_module
from inspect import Parameter, Signature
from typing import Any, Callable, Optional

from ._error import GError
from ._gi import (
    VFuncInfo,
    FunctionInfo,
    CallableInfo,
    CallbackInfo,
    Direction,
    GType,
    TypeInfo,
    TypeTag,
)

__all__ = ["generate_signature"]


tag_pytype = {
    TypeTag.BOOLEAN: bool,
    TypeTag.INT8: int,
    TypeTag.UINT8: int,
    TypeTag.INT16: int,
    TypeTag.UINT16: int,
    TypeTag.INT32: int,
    TypeTag.UINT32: int,
    TypeTag.INT64: int,
    TypeTag.UINT64: int,
    TypeTag.FLOAT: float,
    TypeTag.DOUBLE: float,
    TypeTag.GTYPE: GType,
    TypeTag.UTF8: str,
    TypeTag.FILENAME: str,
    TypeTag.UNICHAR: str,
    TypeTag.ERROR: GError,
}

list_tag_types = {TypeTag.GLIST, TypeTag.GSLIST, TypeTag.ARRAY}


def get_pytype(gi_type: TypeInfo) -> object:
    tag = gi_type.get_tag()
    if pytype := tag_pytype.get(tag):
        return pytype
    if tag == TypeTag.VOID:
        if gi_type.is_pointer():
            return Any
        return None
    if tag in list_tag_types:
        value_type = get_pytype(gi_type.get_param_type(0))
        if value_type is Parameter.empty:
            return list
        return list[value_type]
    if tag == TypeTag.GHASH:
        key_type = get_pytype(gi_type.get_param_type(0))
        value_type = get_pytype(gi_type.get_param_type(1))
        if key_type is Parameter.empty or value_type is Parameter.empty:
            return dict
        return dict[key_type, value_type]
    if tag == TypeTag.INTERFACE:
        info = gi_type.get_interface()
        if isinstance(info, CallbackInfo):
            sig = generate_signature(info)
            return Callable[
                [param.annotation for param in sig.parameters.values()],
                sig.return_annotation,
            ]
        info_name = info.get_name()
        if not info_name:
            return gi_type.get_tag_as_string()
        info_namespace = info.get_namespace()
        module = import_module(f"gi.repository.{info_namespace}")
        try:
            return getattr(module, info_name)
        except NotImplementedError:
            return f"{info_namespace}.{info_name}"
    else:
        return gi_type.get_tag_as_string()


def generate_signature(info: CallableInfo) -> Signature:
    args = info.get_arguments()
    arg_names = {arg.get_name() for arg in args}

    def unique(name):
        while name in arg_names:
            name += "_"
        arg_names.add(name)
        return name

    params = []
    if isinstance(info, FunctionInfo):
        if info.is_constructor():
            params.append(Parameter(unique("cls"), Parameter.POSITIONAL_OR_KEYWORD))
        elif info.is_method():
            params.append(Parameter(unique("self"), Parameter.POSITIONAL_OR_KEYWORD))
    elif isinstance(info, VFuncInfo):
        params.append(Parameter(unique("type"), Parameter.POSITIONAL_OR_KEYWORD))
        params.append(Parameter(unique("self"), Parameter.POSITIONAL_OR_KEYWORD))

    # Build lists of indices prior to adding the docs because it is possible
    # the index retrieved comes before input arguments being used.
    ignore_indices = {info.get_return_type().get_array_length_index()}
    user_data_indices = set()
    for arg in args:
        ignore_indices.add(arg.get_destroy_index())
        ignore_indices.add(arg.get_type_info().get_array_length_index())
        user_data_indices.add(arg.get_closure_index())

    for i, arg in enumerate(args):
        if arg.get_direction() == Direction.OUT:
            continue  # skip exclusively output args
        if i in ignore_indices:
            continue

        default = Parameter.empty
        annotation = get_pytype(arg.get_type_info())
        if arg.may_be_null() or i in user_data_indices:
            # allow-none or user_data from a closure
            default = None
            if annotation is not Parameter.empty and annotation is not Any:
                annotation = Optional[annotation]
        elif arg.is_optional():
            # TODO: Can we retrieve the default value?
            default = ...

        params.append(
            Parameter(
                arg.get_name(),
                Parameter.POSITIONAL_OR_KEYWORD,
                default=default,
                annotation=annotation,
            )
        )

    # Remove defaults from params after the last required parameter.
    last_required = max(
        (i for (i, param) in enumerate(params) if param.default is Parameter.empty),
        default=0,
    )
    for i, param in enumerate(params):
        if i >= last_required:
            break
        if param.default is not Parameter.empty:
            params[i] = param.replace(default=Parameter.empty)

    return_annotation = get_pytype(info.get_return_type())
    if info.may_return_null():
        return_annotation = Optional[return_annotation]

    out_args = []
    for i, arg in enumerate(args):
        if arg.get_direction() == Direction.IN:
            continue  # skip exclusively input args
        if i in ignore_indices:
            continue
        annotation = get_pytype(arg.get_type_info())
        if arg.may_be_null() and annotation is not Parameter.empty:
            annotation = Optional[annotation]
        out_args.append(annotation)

    if return_annotation is not None:
        out_args.insert(0, return_annotation)
    if len(out_args) > 1:
        return_annotation = tuple[tuple(out_args)]
    elif len(out_args) == 1:
        return_annotation = out_args[0]

    return Signature(params, return_annotation=return_annotation)



# pygobject - Python bindings for the GObject library
# Copyright (C) 2006-2007 Johan Dahlin
#
#   gi/_constants.py: GObject type constants
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, see <http://www.gnu.org/licenses/>.

from . import _gi

TYPE_INVALID = _gi.TYPE_INVALID
TYPE_NONE = _gi.GType.from_name("void")
TYPE_INTERFACE = _gi.GType.from_name("GInterface")
TYPE_CHAR = _gi.GType.from_name("gchar")
TYPE_UCHAR = _gi.GType.from_name("guchar")
TYPE_BOOLEAN = _gi.GType.from_name("gboolean")
TYPE_INT = _gi.GType.from_name("gint")
TYPE_UINT = _gi.GType.from_name("guint")
TYPE_LONG = _gi.GType.from_name("glong")
TYPE_ULONG = _gi.GType.from_name("gulong")
TYPE_INT64 = _gi.GType.from_name("gint64")
TYPE_UINT64 = _gi.GType.from_name("guint64")
TYPE_ENUM = _gi.GType.from_name("GEnum")
TYPE_FLAGS = _gi.GType.from_name("GFlags")
TYPE_FLOAT = _gi.GType.from_name("gfloat")
TYPE_DOUBLE = _gi.GType.from_name("gdouble")
TYPE_STRING = _gi.GType.from_name("gchararray")
TYPE_POINTER = _gi.GType.from_name("gpointer")
TYPE_BOXED = _gi.GType.from_name("GBoxed")
TYPE_PARAM = _gi.GType.from_name("GParam")
TYPE_OBJECT = _gi.GType.from_name("GObject")
TYPE_PYOBJECT = _gi.GType.from_name("PyObject")
TYPE_GTYPE = _gi.GType.from_name("GType")
TYPE_STRV = _gi.GType.from_name("GStrv")
TYPE_VARIANT = _gi.GType.from_name("GVariant")
TYPE_UNICHAR = TYPE_UINT



# Copyright 2015 Dustin Spicuzza <dustin@virtualroadside.com>
#           2018 Nikita Churaev <lamefun.x0r@gmail.com>
#           2018 Christoph Reiter <reiter.christoph@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import os
from collections import abc
from functools import partial

from gi.repository import GLib, GObject, Gio


def _extract_handler_and_args(obj_or_map, handler_name):
    handler = None
    if isinstance(obj_or_map, abc.Mapping):
        handler = obj_or_map.get(handler_name, None)
    else:
        handler = getattr(obj_or_map, handler_name, None)

    if handler is None:
        raise AttributeError(f"Handler {handler_name} not found")

    args = ()
    if isinstance(handler, abc.Sequence):
        if len(handler) == 0:
            raise TypeError(f"Handler {handler} tuple can not be empty")
        args = handler[1:]
        handler = handler[0]

    elif not callable(handler):
        raise TypeError(f"Handler {handler} is not a method, function or tuple")

    return handler, args


def define_builder_scope():
    from gi.repository import Gtk

    class BuilderScope(GObject.GObject, Gtk.BuilderScope):
        def __init__(self, scope_object=None):
            super().__init__()
            self._scope_object = scope_object

        def do_create_closure(self, builder, func_name, flags, obj):
            current_object = builder.get_current_object() or self._scope_object

            if not self._scope_object:
                current_object = builder.get_current_object()
                if func_name not in current_object.__gtktemplate_methods__:
                    return None

                current_object.__gtktemplate_handlers__.add(func_name)
                handler_name = current_object.__gtktemplate_methods__[func_name]
            else:
                current_object = self._scope_object
                handler_name = func_name

            swapped = int(flags & Gtk.BuilderClosureFlags.SWAPPED)
            if swapped:
                raise RuntimeError(f"{GObject.ConnectFlags.SWAPPED!r} not supported")
                return None

            handler, args = _extract_handler_and_args(current_object, handler_name)

            if obj:
                p = partial(handler, *args, swap_data=obj)
            else:
                p = partial(handler, *args)

            p.__gtk_template__ = True
            return p

    return BuilderScope


def connect_func(builder, obj, signal_name, handler_name, connect_object, flags, cls):
    if handler_name not in cls.__gtktemplate_methods__:
        return

    method_name = cls.__gtktemplate_methods__[handler_name]
    template_inst = builder.get_object(cls.__gtype_name__)
    template_inst.__gtktemplate_handlers__.add(handler_name)
    handler = getattr(template_inst, method_name)

    after = int(flags & GObject.ConnectFlags.AFTER)
    swapped = int(flags & GObject.ConnectFlags.SWAPPED)
    if swapped:
        raise RuntimeError(f"{GObject.ConnectFlags.SWAPPED!r} not supported")

    if connect_object is not None:
        func = obj.connect_object_after if after else obj.connect_object
        func(signal_name, handler, connect_object)
    else:
        func = obj.connect_after if after else obj.connect
        func(signal_name, handler)


def register_template(cls):
    from gi.repository import Gtk

    bound_methods = {}
    bound_widgets = {}

    for attr_name, obj in list(cls.__dict__.items()):
        if isinstance(obj, CallThing):
            setattr(cls, attr_name, obj._func)
            handler_name = obj._name
            if handler_name is None:
                handler_name = attr_name

            if handler_name in bound_methods:
                old_attr_name = bound_methods[handler_name]
                raise RuntimeError(
                    f"Error while exposing handler {handler_name!r} as {attr_name!r}, "
                    f"already available as {old_attr_name!r}"
                )
            bound_methods[handler_name] = attr_name
        elif isinstance(obj, Child):
            widget_name = obj._name
            if widget_name is None:
                widget_name = attr_name

            if widget_name in bound_widgets:
                old_attr_name = bound_widgets[widget_name]
                raise RuntimeError(
                    f"Error while exposing child {widget_name!r} as {attr_name!r}, "
                    f"already available as {old_attr_name!r}"
                )
            bound_widgets[widget_name] = attr_name
            cls.bind_template_child_full(widget_name, obj._internal, 0)

    cls.__gtktemplate_methods__ = bound_methods
    cls.__gtktemplate_widgets__ = bound_widgets

    if Gtk._version == "4.0":
        BuilderScope = define_builder_scope()
        cls.set_template_scope(BuilderScope())
    else:
        cls.set_connect_func(connect_func, cls)

    base_init_template = cls.init_template
    cls.__dontuse_ginstance_init__ = lambda s: init_template(s, cls, base_init_template)
    # To make this file work with older PyGObject we expose our init code
    # as init_template() but make it a noop when we call it ourselves first
    cls.init_template = cls.__dontuse_ginstance_init__


def init_template(self, cls, base_init_template):
    self.init_template = lambda: None

    if self.__class__ is not cls:
        raise TypeError(
            "Inheritance from classes with @Gtk.Template decorators "
            "is not allowed at this time"
        )

    self.__gtktemplate_handlers__ = set()

    base_init_template(self)

    for widget_name, attr_name in self.__gtktemplate_widgets__.items():
        self.__dict__[attr_name] = self.get_template_child(cls, widget_name)

    for handler_name, attr_name in self.__gtktemplate_methods__.items():
        if handler_name not in self.__gtktemplate_handlers__:
            raise RuntimeError(
                f"Handler '{handler_name}' was declared with @Gtk.Template.Callback "
                "but was not present in template"
            )


class Child:
    def __init__(self, name=None, **kwargs):
        self._name = name
        self._internal = kwargs.pop("internal", False)
        if kwargs:
            raise TypeError(f"Unhandled arguments: {kwargs!r}")


class CallThing:
    def __init__(self, name, func):
        self._name = name
        self._func = func


class Callback:
    def __init__(self, name=None):
        self._name = name

    def __call__(self, func):
        return CallThing(self._name, func)


def validate_resource_path(path):
    """Raises GLib.Error in case the resource doesn't exist."""
    try:
        Gio.resources_get_info(path, Gio.ResourceLookupFlags.NONE)
    except GLib.Error:
        # resources_get_info() doesn't handle overlays but we keep using it
        # as a fast path.
        # https://gitlab.gnome.org/GNOME/pygobject/issues/230
        Gio.resources_lookup_data(path, Gio.ResourceLookupFlags.NONE)


class Template:
    def __init__(self, **kwargs):
        self.string = None
        self.filename = None
        self.resource_path = None
        if "string" in kwargs:
            self.string = kwargs.pop("string")
        elif "filename" in kwargs:
            self.filename = kwargs.pop("filename")
        elif "resource_path" in kwargs:
            self.resource_path = kwargs.pop("resource_path")
        else:
            raise TypeError(
                "Requires one of the following arguments: "
                "string, filename, resource_path"
            )

        if kwargs:
            raise TypeError(f"Unhandled keyword arguments {kwargs!r}")

    @classmethod
    def from_file(cls, filename):
        return cls(filename=filename)

    @classmethod
    def from_string(cls, string):
        return cls(string=string)

    @classmethod
    def from_resource(cls, resource_path):
        return cls(resource_path=resource_path)

    Callback = Callback

    Child = Child

    def __call__(self, cls):
        from gi.repository import Gtk

        if not isinstance(cls, type) or not issubclass(cls, Gtk.Widget):
            raise TypeError("Can only use @Gtk.Template on Widgets")

        if "__gtype_name__" not in cls.__dict__:
            raise TypeError(
                f"{cls.__name__!r} does not have a __gtype_name__. Set it to the name "
                "of the class in your template"
            )

        if hasattr(cls, "__gtktemplate_methods__"):
            raise TypeError("Cannot nest template classes")

        if self.string is not None:
            data = self.string
            if not isinstance(data, bytes):
                data = data.encode("utf-8")
            bytes_ = GLib.Bytes.new(data)
            cls.set_template(bytes_)
            register_template(cls)
            return cls
        if self.resource_path is not None:
            validate_resource_path(self.resource_path)
            cls.set_template_from_resource(self.resource_path)
            register_template(cls)
            return cls
        assert self.filename is not None
        file_ = Gio.File.new_for_path(os.fspath(self.filename))
        bytes_ = GLib.Bytes.new(file_.load_contents()[1])
        cls.set_template(bytes_)
        register_template(cls)
        return cls


__all__ = ["Template"]



# Copyright (C) 2010 Tomeu Vizoso <tomeu.vizoso@collabora.co.uk>
# Copyright (C) 2011, 2012 Canonical Ltd.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA
import contextlib
import warnings
import sys
import socket

from .._ossighelper import register_sigint_fallback, get_event_loop
from ..module import get_introspection_module
from .._gi import (
    variant_type_from_string,
    source_new,
    source_set_callback,
    io_channel_read,
    main_context_query,
)
from ..overrides import override, deprecated, deprecated_attr
from gi import PyGIDeprecationWarning, version_info

GLib = get_introspection_module("GLib")

__all__ = []

from gi import _option as option

option
__all__.append("option")


# Types and functions still needed from static bindings
from gi import _gi
from gi._error import GError

Error = GError
OptionContext = _gi.OptionContext
OptionGroup = _gi.OptionGroup
Pid = _gi.Pid
spawn_async = _gi.spawn_async


def threads_init():
    warnings.warn(
        "Since version 3.11, calling threads_init is no longer needed. "
        "See: https://pygobject.gnome.org/guide/threading.html#threads-faq",
        PyGIDeprecationWarning,
        stacklevel=2,
    )


def gerror_matches(self, domain, code):
    # Handle cases where self.domain was set to an integer for compatibility
    # with the introspected GLib.Error.
    if isinstance(self.domain, str):
        self_domain_quark = GLib.quark_from_string(self.domain)
    else:
        self_domain_quark = self.domain
    return (self_domain_quark, self.code) == (domain, code)


def gerror_new_literal(domain, message, code):
    domain_quark = GLib.quark_to_string(domain)
    return GError(message, domain_quark, code)


# Monkey patch methods that rely on GLib introspection to be loaded at runtime.
Error.__name__ = "Error"
Error.__module__ = "gi.repository.GLib"
Error.__gtype__ = GLib.Error.__gtype__
Error.matches = gerror_matches
Error.new_literal = staticmethod(gerror_new_literal)


__all__ += [
    "Error",
    "GError",
    "OptionContext",
    "OptionGroup",
    "Pid",
    "spawn_async",
    "threads_init",
]


class _VariantCreator:
    _LEAF_CONSTRUCTORS = {
        "b": GLib.Variant.new_boolean,
        "y": GLib.Variant.new_byte,
        "n": GLib.Variant.new_int16,
        "q": GLib.Variant.new_uint16,
        "i": GLib.Variant.new_int32,
        "u": GLib.Variant.new_uint32,
        "x": GLib.Variant.new_int64,
        "t": GLib.Variant.new_uint64,
        "h": GLib.Variant.new_handle,
        "d": GLib.Variant.new_double,
        "s": GLib.Variant.new_string,
        "o": GLib.Variant.new_object_path,
        "g": GLib.Variant.new_signature,
        "v": GLib.Variant.new_variant,
    }

    def _create(self, format, value):
        """Create a GVariant object from given format and a value that matches
        the format.

        This method recursively calls itself for complex structures (arrays,
        dictionaries, boxed).

        Returns the generated GVariant.

        If value is None it will generate an empty GVariant container type.
        """
        gvtype = GLib.VariantType(format)
        if format in self._LEAF_CONSTRUCTORS:
            return self._LEAF_CONSTRUCTORS[format](value)

        # Since we discarded all leaf types, this must be a container
        builder = GLib.VariantBuilder.new(gvtype)
        if value is None:
            return builder.end()

        if gvtype.is_maybe():
            builder.add_value(self._create(gvtype.element().dup_string(), value))
            return builder.end()

        try:
            iter(value)
        except TypeError:
            raise TypeError(
                f"Could not create array, tuple or dictionary entry from non iterable value {format} {value}"
            )

        if gvtype.is_tuple() and gvtype.n_items() != len(value):
            raise TypeError(
                f"Tuple mismatches value's number of elements {format} {value}"
            )
        if gvtype.is_dict_entry() and len(value) != 2:
            raise TypeError(
                f"Dictionary entries must have two elements {format} {value}"
            )

        if gvtype.is_array():
            element_type = gvtype.element().dup_string()
            if isinstance(value, dict):
                value = value.items()
            for i in value:
                builder.add_value(self._create(element_type, i))
        else:
            remainer_format = format[1:]
            for i in value:
                dup = variant_type_from_string(remainer_format).dup_string()
                builder.add_value(self._create(dup, i))
                remainer_format = remainer_format[len(dup) :]

        return builder.end()


LEAF_ACCESSORS = {
    "b": "get_boolean",
    "y": "get_byte",
    "n": "get_int16",
    "q": "get_uint16",
    "i": "get_int32",
    "u": "get_uint32",
    "x": "get_int64",
    "t": "get_uint64",
    "h": "get_handle",
    "d": "get_double",
    "s": "get_string",
    "o": "get_string",  # object path
    "g": "get_string",  # signature
}


class Variant(GLib.Variant):
    def __new__(cls, format_string, value):
        """Create a GVariant from a native Python object.

        format_string is a standard GVariant type signature, value is a Python
        object whose structure has to match the signature.

        Examples:
          GLib.Variant('i', 1)
          GLib.Variant('(is)', (1, 'hello'))
          GLib.Variant('(asa{sv})', ([], {'foo': GLib.Variant('b', True),
                                          'bar': GLib.Variant('i', 2)}))

        """
        if not GLib.VariantType.string_is_valid(format_string):
            raise TypeError("Invalid GVariant format string '%s'", format_string)
        creator = _VariantCreator()
        v = creator._create(format_string, value)
        v.format_string = format_string
        return v

    @staticmethod
    def new_tuple(*elements):
        return GLib.Variant.new_tuple(elements)

    def __del__(self):
        with contextlib.suppress(ImportError):
            # Calling unref will cause gi and gi.repository.GLib to be
            # imported. However, if the program is exiting, then these
            # modules have likely been removed from sys.modules and will
            # raise an exception. Assume that's the case for ImportError
            # and ignore the exception since everything will be cleaned
            # up, anyway.
            self.unref()

    def __str__(self):
        return self.print_(True)

    def __repr__(self):
        if hasattr(self, "format_string"):
            f = self.format_string
        else:
            f = self.get_type_string()
        return f"GLib.Variant('{f}', {self.print_(False)})"

    def __eq__(self, other):
        try:
            return self.equal(other)
        except TypeError:
            return False

    def __ne__(self, other):
        try:
            return not self.equal(other)
        except TypeError:
            return True

    def __hash__(self):
        # We're not using just hash(self.unpack()) because otherwise we'll have
        # hash collisions between the same content in different variant types,
        # which will cause a performance issue in set/dict/etc.
        return hash((self.get_type_string(), self.unpack()))

    def unpack(self):
        """Decompose a GVariant into a native Python object."""
        type_string = self.get_type_string()

        # simple values
        la = LEAF_ACCESSORS.get(type_string)
        if la:
            return getattr(self, la)()

        # tuple
        if type_string.startswith("("):
            return tuple(
                self.get_child_value(i).unpack() for i in range(self.n_children())
            )

        # dictionary
        if type_string.startswith("a{"):
            res = {}
            for i in range(self.n_children()):
                v = self.get_child_value(i)
                res[v.get_child_value(0).unpack()] = v.get_child_value(1).unpack()
            return res

        # array
        if type_string.startswith("a"):
            return [self.get_child_value(i).unpack() for i in range(self.n_children())]

        # variant (just unbox transparently)
        if type_string.startswith("v"):
            return self.get_variant().unpack()

        # maybe
        if type_string.startswith("m"):
            if not self.n_children():
                return None
            return self.get_child_value(0).unpack()

        raise NotImplementedError("unsupported GVariant type " + type_string)

    @classmethod
    def split_signature(klass, signature):
        """Return a list of the element signatures of the topmost signature tuple.

        If the signature is not a tuple, it returns one element with the entire
        signature. If the signature is an empty tuple, the result is [].

        This is useful for e. g. iterating over method parameters which are
        passed as a single Variant.
        """
        if signature == "()":
            return []

        if not signature.startswith("("):
            return [signature]

        result = []
        head = ""
        tail = signature[1:-1]  # eat the surrounding ()
        while tail:
            c = tail[0]
            head += c
            tail = tail[1:]

            if c in ("m", "a"):
                # prefixes, keep collecting
                continue
            if c in ("(", "{"):
                # consume until corresponding )/}
                level = 1
                up = c
                down = ")" if up == "(" else "}"
                while level > 0:
                    c = tail[0]
                    head += c
                    tail = tail[1:]
                    if c == up:
                        level += 1
                    elif c == down:
                        level -= 1

            # otherwise we have a simple type
            result.append(head)
            head = ""

        return result

    #
    # Pythonic iterators
    #

    def __len__(self):
        if self.get_type_string() in ["s", "o", "g"]:
            return len(self.get_string())
        # Array, dict, tuple
        if self.get_type_string().startswith("a") or self.get_type_string().startswith(
            "("
        ):
            return self.n_children()
        raise TypeError(
            f"GVariant type {self.get_type_string()} does not have a length"
        )

    def __getitem__(self, key):
        # dict
        if self.get_type_string().startswith("a{"):
            try:
                val = self.lookup_value(key, variant_type_from_string("*"))
                if val is None:
                    raise KeyError(key)
                return val.unpack()
            except TypeError:
                # lookup_value() only works for string keys, which is certainly
                # the common case; we have to do painful iteration for other
                # key types
                for i in range(self.n_children()):
                    v = self.get_child_value(i)
                    if v.get_child_value(0).unpack() == key:
                        return v.get_child_value(1).unpack()
                raise KeyError(key)

        # array/tuple
        if self.get_type_string().startswith("a") or self.get_type_string().startswith(
            "("
        ):
            key = int(key)
            if key < 0:
                key = self.n_children() + key
            if key < 0 or key >= self.n_children():
                raise IndexError("list index out of range")
            return self.get_child_value(key).unpack()

        # string
        if self.get_type_string() in ["s", "o", "g"]:
            return self.get_string().__getitem__(key)

        raise TypeError(f"GVariant type {self.get_type_string()} is not a container")

    #
    # Pythonic bool operations
    #

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        if self.get_type_string() in ["y", "n", "q", "i", "u", "x", "t", "h", "d"]:
            return self.unpack() != 0
        if self.get_type_string() in ["b"]:
            return self.get_boolean()
        if self.get_type_string() in ["s", "o", "g"]:
            return len(self.get_string()) != 0
        # Array, dict, tuple
        if self.get_type_string().startswith("a") or self.get_type_string().startswith(
            "("
        ):
            return self.n_children() != 0
        # unpack works recursively, hence bool also works recursively
        return bool(self.unpack())

    def keys(self):
        if not self.get_type_string().startswith("a{"):
            raise TypeError(
                f"GVariant type {self.get_type_string()} is not a dictionary"
            )

        res = []
        for i in range(self.n_children()):
            v = self.get_child_value(i)
            res.append(v.get_child_value(0).unpack())
        return res


def get_string(self):
    value, _length = GLib.Variant.get_string(self)
    return value


setattr(Variant, "get_string", get_string)

__all__.append("Variant")


def markup_escape_text(text, length=-1):
    if isinstance(text, bytes):
        return GLib.markup_escape_text(text.decode("UTF-8"), length)
    return GLib.markup_escape_text(text, length)


__all__.append("markup_escape_text")


# backwards compatible names from old static bindings
for n in [
    "DESKTOP",
    "DOCUMENTS",
    "DOWNLOAD",
    "MUSIC",
    "PICTURES",
    "PUBLIC_SHARE",
    "TEMPLATES",
    "VIDEOS",
]:
    attr = "USER_DIRECTORY_" + n
    deprecated_attr("GLib", attr, "GLib.UserDirectory.DIRECTORY_" + n)
    globals()[attr] = getattr(GLib.UserDirectory, "DIRECTORY_" + n)
    __all__.append(attr)

for n in ["ERR", "HUP", "IN", "NVAL", "OUT", "PRI"]:
    globals()["IO_" + n] = getattr(GLib.IOCondition, n)
    __all__.append("IO_" + n)

for n in [
    "APPEND",
    "GET_MASK",
    "IS_READABLE",
    "IS_SEEKABLE",
    "MASK",
    "NONBLOCK",
    "SET_MASK",
]:
    attr = "IO_FLAG_" + n
    deprecated_attr("GLib", attr, "GLib.IOFlags." + n)
    globals()[attr] = getattr(GLib.IOFlags, n)
    __all__.append(attr)

# spelling for the win
IO_FLAG_IS_WRITEABLE = GLib.IOFlags.IS_WRITABLE
deprecated_attr("GLib", "IO_FLAG_IS_WRITEABLE", "GLib.IOFlags.IS_WRITABLE")
__all__.append("IO_FLAG_IS_WRITEABLE")

for n in ["AGAIN", "EOF", "ERROR", "NORMAL"]:
    attr = "IO_STATUS_" + n
    globals()[attr] = getattr(GLib.IOStatus, n)
    deprecated_attr("GLib", attr, "GLib.IOStatus." + n)
    __all__.append(attr)

for n in [
    "CHILD_INHERITS_STDIN",
    "DO_NOT_REAP_CHILD",
    "FILE_AND_ARGV_ZERO",
    "LEAVE_DESCRIPTORS_OPEN",
    "SEARCH_PATH",
    "STDERR_TO_DEV_NULL",
    "STDOUT_TO_DEV_NULL",
]:
    attr = "SPAWN_" + n
    globals()[attr] = getattr(GLib.SpawnFlags, n)
    deprecated_attr("GLib", attr, "GLib.SpawnFlags." + n)
    __all__.append(attr)

for n in [
    "HIDDEN",
    "IN_MAIN",
    "REVERSE",
    "NO_ARG",
    "FILENAME",
    "OPTIONAL_ARG",
    "NOALIAS",
]:
    attr = "OPTION_FLAG_" + n
    globals()[attr] = getattr(GLib.OptionFlags, n)
    deprecated_attr("GLib", attr, "GLib.OptionFlags." + n)
    __all__.append(attr)

for n in ["UNKNOWN_OPTION", "BAD_VALUE", "FAILED"]:
    attr = "OPTION_ERROR_" + n
    deprecated_attr("GLib", attr, "GLib.OptionError." + n)
    globals()[attr] = getattr(GLib.OptionError, n)
    __all__.append(attr)


# these are not currently exported in GLib gir, presumably because they are
# platform dependent; so get them from our static bindings
for name in [
    "G_MINFLOAT",
    "G_MAXFLOAT",
    "G_MINDOUBLE",
    "G_MAXDOUBLE",
    "G_MINSHORT",
    "G_MAXSHORT",
    "G_MAXUSHORT",
    "G_MININT",
    "G_MAXINT",
    "G_MAXUINT",
    "G_MINLONG",
    "G_MAXLONG",
    "G_MAXULONG",
    "G_MAXSIZE",
    "G_MINSSIZE",
    "G_MAXSSIZE",
    "G_MINOFFSET",
    "G_MAXOFFSET",
]:
    attr = name.split("_", 1)[-1]
    globals()[attr] = getattr(_gi, name)
    __all__.append(attr)


class MainLoop(GLib.MainLoop):
    # Backwards compatible constructor API
    def __new__(cls, context=None):
        return GLib.MainLoop.new(context, False)

    def __init__(self, context=None):
        pass

    def run(self):
        with (
            register_sigint_fallback(self.quit),
            get_event_loop(self.get_context()).running(self.quit),
        ):
            super().run()


MainLoop = override(MainLoop)
__all__.append("MainLoop")


class MainContext(GLib.MainContext):
    # Backwards compatible API with default value
    def iteration(self, may_block=True):
        with get_event_loop(self).paused():
            return super().iteration(may_block)

    def query(self, max_priority: int) -> tuple[int, list[GLib.PollFD]]:
        """Determines information necessary to poll this main loop.

        :param max_priority: maximum priority source to check
        :returns:
          The timeout (msec) used for polling,
          and the list of poll fd's.

        Please also check the usage notes in the
        `C documention <https://docs.gtk.org/glib/method.MainContext.query.html>`__.
        """
        return main_context_query(self, max_priority)


MainContext = override(MainContext)
__all__.append("MainContext")


class Source(GLib.Source):
    def __new__(cls, *args, **kwargs):
        # use our custom pygi_source_new() here as g_source_new() is not
        # bindable
        source = source_new()
        source.__class__ = cls
        setattr(source, "__pygi_custom_source", True)
        return source

    def __init__(self, *args, **kwargs):
        return super().__init__()

    def __del__(self):
        if hasattr(self, "__pygi_custom_source"):
            # We destroy and finalize the box from here, as GLib might hold
            # a reference (e.g. while the source is pending), delaying the
            # finalize call until a later point.
            self.destroy()
            self.finalize()
            self._clear_boxed()

    def finalize(self):
        pass

    def set_callback(self, fn, user_data=None):
        if hasattr(self, "__pygi_custom_source"):
            # use our custom pygi_source_set_callback() if for a GSource object
            # with custom functions
            source_set_callback(self, fn, user_data)
        else:
            # otherwise, for Idle and Timeout, use the standard method
            super().set_callback(fn, user_data)

    def get_current_time(self):
        return GLib.get_real_time() * 0.000001

    get_current_time = deprecated(
        get_current_time, "GLib.Source.get_time() or GLib.get_real_time()"
    )

    # as get/set_priority are introspected, we can't use the static
    # property(get_priority, ..) here
    def __get_priority(self):
        return self.get_priority()

    def __set_priority(self, value):
        self.set_priority(value)

    priority = property(__get_priority, __set_priority)

    def __get_can_recurse(self):
        return self.get_can_recurse()

    def __set_can_recurse(self, value):
        self.set_can_recurse(value)

    can_recurse = property(__get_can_recurse, __set_can_recurse)


Source = override(Source)
__all__.append("Source")


class Idle(Source):
    def __new__(cls, priority=GLib.PRIORITY_DEFAULT):
        source = GLib.idle_source_new()
        source.__class__ = cls
        return source

    def __init__(self, priority=GLib.PRIORITY_DEFAULT):
        super(Source, self).__init__()
        if priority != GLib.PRIORITY_DEFAULT:
            self.set_priority(priority)


__all__.append("Idle")


class Timeout(Source):
    def __new__(cls, interval=0, priority=GLib.PRIORITY_DEFAULT):
        source = GLib.timeout_source_new(interval)
        source.__class__ = cls
        return source

    def __init__(self, interval=0, priority=GLib.PRIORITY_DEFAULT):
        if priority != GLib.PRIORITY_DEFAULT:
            self.set_priority(priority)


__all__.append("Timeout")


# backwards compatible API
def idle_add(function, *user_data, priority=GLib.PRIORITY_DEFAULT_IDLE):
    return GLib.idle_add(priority, function, *user_data)


__all__.append("idle_add")


def timeout_add(interval, function, *user_data, priority=GLib.PRIORITY_DEFAULT):
    return GLib.timeout_add(priority, interval, function, *user_data)


__all__.append("timeout_add")


def timeout_add_seconds(interval, function, *user_data, priority=GLib.PRIORITY_DEFAULT):
    return GLib.timeout_add_seconds(priority, interval, function, *user_data)


__all__.append("timeout_add_seconds")


# The GI GLib API uses g_io_add_watch_full renamed to g_io_add_watch with
# a signature of (channel, priority, condition, func, user_data).
# Prior to PyGObject 3.8, this function was statically bound with an API closer to the
# non-full version with a signature of: (fd, condition, func, *user_data)
# We need to support this until we are okay with breaking API in a way which is
# not backwards compatible.
#
# This needs to take into account several historical APIs:
# - calling with an fd as first argument
# - calling with a Python file object as first argument (we keep this one as
#   it's really convenient and does not change the number of arguments)
# - calling without a priority as second argument
def _io_add_watch_get_args(channel, priority_, condition, *cb_and_user_data, **kwargs):
    if not isinstance(priority_, int) or isinstance(priority_, GLib.IOCondition):
        warnings.warn(
            "Calling io_add_watch without priority as second argument is deprecated",
            PyGIDeprecationWarning,
        )
        # shift the arguments around
        user_data = cb_and_user_data
        callback = condition
        condition = priority_
        if not callable(callback):
            raise TypeError("third argument must be callable")

        # backwards compatibility: Call with priority kwarg
        if "priority" in kwargs:
            warnings.warn(
                "Calling io_add_watch with priority keyword argument is deprecated, put it as second positional argument",
                PyGIDeprecationWarning,
            )
            priority_ = kwargs["priority"]
        else:
            priority_ = GLib.PRIORITY_DEFAULT
    else:
        if len(cb_and_user_data) < 1 or not callable(cb_and_user_data[0]):
            raise TypeError("expecting callback as fourth argument")
        callback = cb_and_user_data[0]
        user_data = cb_and_user_data[1:]

    # backwards compatibility: Allow calling with fd
    if isinstance(channel, int):

        def func_fdtransform(_, cond, *data):
            return callback(channel, cond, *data)

        real_channel = GLib.IOChannel.unix_new(channel)
    elif isinstance(channel, socket.socket) and sys.platform == "win32":

        def func_fdtransform(_, cond, *data):
            return callback(channel, cond, *data)

        real_channel = GLib.IOChannel.win32_new_socket(channel.fileno())
    elif hasattr(channel, "fileno"):
        # backwards compatibility: Allow calling with Python file
        def func_fdtransform(_, cond, *data):
            return callback(channel, cond, *data)

        real_channel = GLib.IOChannel.unix_new(channel.fileno())
    else:
        assert isinstance(channel, GLib.IOChannel)
        func_fdtransform = callback
        real_channel = channel

    return real_channel, priority_, condition, func_fdtransform, user_data


__all__.append("_io_add_watch_get_args")


def io_add_watch(*args, **kwargs):
    """io_add_watch(channel, priority, condition, func, *user_data) -> event_source_id."""
    channel, priority, condition, func, user_data = _io_add_watch_get_args(
        *args, **kwargs
    )
    return GLib.io_add_watch(channel, priority, condition, func, *user_data)


__all__.append("io_add_watch")


# backwards compatible API
class IOChannel(GLib.IOChannel):
    def __new__(cls, filedes=None, filename=None, mode=None, hwnd=None):
        if filedes is not None:
            return GLib.IOChannel.unix_new(filedes)
        if filename is not None:
            return GLib.IOChannel.new_file(filename, mode or "r")
        if hwnd is not None:
            return GLib.IOChannel.win32_new_fd(hwnd)
        raise TypeError(
            "either a valid file descriptor, file name, or window handle must be supplied"
        )

    def __init__(self, *args, **kwargs):
        return super().__init__()

    def read(self, max_count=-1):
        """Reads data from a :obj:`~gi.repository.GLib.IOChannel`."""
        return io_channel_read(self, max_count)

    def read_chars(self, max_count=-1):
        """Alias for GLib.IOChannel.read()."""
        return self.read(max_count)

    def readline(self, size_hint=-1):
        # note, size_hint is just to maintain backwards compatible API; the
        # old static binding did not actually use it
        (_status, buf, _length, _terminator_pos) = self.read_line()
        if buf is None:
            return ""
        return buf

    def readlines(self, size_hint=-1):
        # note, size_hint is just to maintain backwards compatible API;
        # the old static binding did not actually use it
        lines = []
        status = GLib.IOStatus.NORMAL
        while status == GLib.IOStatus.NORMAL:
            (status, buf, _length, _terminator_pos) = self.read_line()
            # note, this appends an empty line after EOF; this is
            # bug-compatible with the old static bindings
            if buf is None:
                buf = ""
            lines.append(buf)
        return lines

    def write(self, buf, buflen=-1):
        if not isinstance(buf, bytes):
            buf = buf.encode("UTF-8")
        if buflen == -1:
            buflen = len(buf)
        (_status, written) = self.write_chars(buf, buflen)
        return written

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    _whence_map = {0: GLib.SeekType.SET, 1: GLib.SeekType.CUR, 2: GLib.SeekType.END}

    def seek(self, offset, whence=0):
        try:
            w = self._whence_map[whence]
        except KeyError:
            raise ValueError("invalid 'whence' value")
        return self.seek_position(offset, w)

    def add_watch(
        self, condition, callback, *user_data, priority=GLib.PRIORITY_DEFAULT
    ):
        return io_add_watch(self, priority, condition, callback, *user_data)

    add_watch = deprecated(add_watch, "GLib.io_add_watch()")

    def __iter__(self):
        return self

    def __next__(self):
        (status, buf, _length, _terminator_pos) = self.read_line()
        if status == GLib.IOStatus.NORMAL:
            return buf
        raise StopIteration


IOChannel = override(IOChannel)
__all__.append("IOChannel")


class PollFD(GLib.PollFD):
    def __new__(cls, fd, events):
        pollfd = GLib.PollFD()
        pollfd.__class__ = cls
        return pollfd

    def __init__(self, fd, events):
        self.fd = fd
        self.events = events


PollFD = override(PollFD)
__all__.append("PollFD")


# The GI GLib API uses g_child_watch_add_full renamed to g_child_watch_add with
# a signature of (priority, pid, callback, data).
# Prior to PyGObject 3.8, this function was statically bound with an API closer to the
# non-full version with a signature of: (pid, callback, data=None, priority=GLib.PRIORITY_DEFAULT)
# We need to support this until we are okay with breaking API in a way which is
# not backwards compatible.
def _child_watch_add_get_args(priority_or_pid, pid_or_callback, *args, **kwargs):
    user_data = []

    if callable(pid_or_callback):
        warnings.warn(
            "Calling child_watch_add without priority as first argument is deprecated",
            PyGIDeprecationWarning,
        )
        pid = priority_or_pid
        callback = pid_or_callback
        if len(args) == 0:
            priority = kwargs.get("priority", GLib.PRIORITY_DEFAULT)
        elif len(args) == 1:
            user_data = args
            priority = kwargs.get("priority", GLib.PRIORITY_DEFAULT)
        elif len(args) == 2:
            user_data = [args[0]]
            priority = args[1]
        else:
            raise TypeError("expected at most 4 positional arguments")
    else:
        priority = priority_or_pid
        pid = pid_or_callback
        if "function" in kwargs:
            callback = kwargs["function"]
            user_data = args
        elif len(args) > 0 and callable(args[0]):
            callback = args[0]
            user_data = args[1:]
        else:
            raise TypeError("expected callback as third argument")

    if "data" in kwargs:
        if user_data:
            raise TypeError('got multiple values for "data" argument')
        user_data = (kwargs["data"],)

    return priority, pid, callback, user_data


# we need this to be accessible for unit testing
__all__.append("_child_watch_add_get_args")


def child_watch_add(*args, **kwargs):
    """child_watch_add(priority, pid, function, *data)."""
    priority, pid, function, data = _child_watch_add_get_args(*args, **kwargs)
    return GLib.child_watch_add(priority, pid, function, *data)


__all__.append("child_watch_add")


def get_current_time():
    return GLib.get_real_time() * 0.000001


get_current_time = deprecated(get_current_time, "GLib.get_real_time()")

__all__.append("get_current_time")


# backwards compatible API with default argument, and ignoring bytes_read
# output argument
def filename_from_utf8(utf8string, len=-1):
    return GLib.filename_from_utf8(utf8string, len)[0]


__all__.append("filename_from_utf8")


if hasattr(GLib, "unix_signal_add"):
    unix_signal_add_full = GLib.unix_signal_add
    __all__.append("unix_signal_add_full")
    deprecated_attr("GLib", "unix_signal_add_full", "GLib.unix_signal_add")


# obsolete constants for backwards compatibility
glib_version = (GLib.MAJOR_VERSION, GLib.MINOR_VERSION, GLib.MICRO_VERSION)
__all__.append("glib_version")
deprecated_attr(
    "GLib",
    "glib_version",
    "(GLib.MAJOR_VERSION, GLib.MINOR_VERSION, GLib.MICRO_VERSION)",
)

pyglib_version = version_info
__all__.append("pyglib_version")
deprecated_attr("GLib", "pyglib_version", "gi.version_info")



# Copyright (C) 2005-2009 Johan Dahlin <johan@gnome.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

# support overrides in different directories than our gi module
from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

import sys
import os
import importlib
import types

_static_binding_error = (
    "When using gi.repository you must not import static "
    'modules like "gobject". Please change all occurrences '
    'of "import gobject" to "from gi.repository import GObject". '
    "See: https://bugzilla.gnome.org/show_bug.cgi?id=709183"
)

# we can't have pygobject 2 loaded at the same time we load the internal _gobject
if "gobject" in sys.modules:
    raise ImportError(_static_binding_error)


from . import _gi
from ._gi import _API as _API
from ._gi import Repository
from ._gi import PyGIDeprecationWarning as PyGIDeprecationWarning
from ._gi import PyGIWarning as PyGIWarning

_versions = {}
_overridesdir = os.path.join(os.path.dirname(__file__), "overrides")

# Needed for compatibility with "pygobject.h"/pygobject_init()
_gobject = types.ModuleType("gi._gobject")
sys.modules[_gobject.__name__] = _gobject
_gobject._PyGObject_API = _gi._PyGObject_API
_gobject.pygobject_version = _gi.pygobject_version

version_info = _gi.pygobject_version[:]
__version__ = "{}.{}.{}".format(*version_info)

_gi.register_foreign()

# Options which can be enabled or disabled after importing 'gi'. This may affect
# repository imports or binding machinery in a backwards incompatible way.
_options = {
    # When True, importing Gtk or Gdk will call Gtk.init() or Gdk.init() respectively.
    "legacy_autoinit": True,
}


class _DummyStaticModule(types.ModuleType):
    __path__ = None

    def __getattr__(self, name):
        raise AttributeError(_static_binding_error)


sys.modules["glib"] = _DummyStaticModule("glib", _static_binding_error)
sys.modules["gobject"] = _DummyStaticModule("gobject", _static_binding_error)
sys.modules["gio"] = _DummyStaticModule("gio", _static_binding_error)
sys.modules["gtk"] = _DummyStaticModule("gtk", _static_binding_error)
sys.modules["gtk.gdk"] = _DummyStaticModule("gtk.gdk", _static_binding_error)


def check_version(version):
    if isinstance(version, str):
        version_list = tuple(map(int, version.split(".")))
    else:
        version_list = version

    if version_list > version_info:
        raise ValueError(
            f"pygobject's version {version} required, and available version "
            f"{__version__} is not recent enough"
        )


def require_version(namespace, version):
    """Ensures the correct versions are loaded when importing `gi` modules.

    :param namespace: The name of module to require.
    :type namespace: str
    :param version: The version of module to require.
    :type version: str
    :raises ValueError: If module/version is already loaded, already required, or unavailable.

    :Example:

    .. code-block:: python

        import gi

        gi.require_version("Gtk", "3.0")

    """
    repository = Repository.get_default()

    if not isinstance(version, str):
        raise ValueError("Namespace version needs to be a string.")

    if namespace in repository.get_loaded_namespaces():
        loaded_version = repository.get_version(namespace)
        if loaded_version != version:
            raise ValueError(
                f"Namespace {namespace} is already loaded with version {loaded_version}"
            )

    if namespace in _versions and _versions[namespace] != version:
        raise ValueError(
            f"Namespace {namespace} already requires version {_versions[namespace]}"
        )

    available_versions = repository.enumerate_versions(namespace)
    if not available_versions:
        raise ValueError(f"Namespace {namespace} not available")

    if version not in available_versions:
        raise ValueError(f"Namespace {namespace} not available for version {version}")

    _versions[namespace] = version


def require_versions(requires):
    """Utility function for consolidating multiple `gi.require_version()` calls.

    :param requires: The names and versions of modules to require.
    :type requires: dict

    :Example:

    .. code-block:: python

        import gi

        gi.require_versions({"Gtk": "3.0", "GLib": "2.0", "Gio": "2.0"})
    """
    for module_name, module_version in requires.items():
        require_version(module_name, module_version)


def get_required_version(namespace):
    return _versions.get(namespace)


def require_foreign(namespace, symbol=None):
    """Ensure the given foreign marshaling module is available and loaded.

    :param str namespace:
        Introspection namespace of the foreign module (e.g. "cairo")
    :param symbol:
        Optional symbol typename to ensure a converter exists.
    :type symbol: str or None
    :raises: ImportError

    :Example:

    .. code-block:: python

        import gi
        import cairo

        gi.require_foreign("cairo")

    """
    try:
        _gi.require_foreign(namespace, symbol)
    except Exception as e:
        raise ImportError(str(e))
    importlib.import_module("gi.repository", namespace)


def get_option(name):
    """Get option by name.

    :param str name:
        Name of the option.
    """
    return _options.get(name)


def disable_legacy_autoinit():
    """Disable the legacy initialization of Gdk and Gtk.

    These modules are automatically initialized on import from `gi.repository`.
    This is not required, and was only kept for backwards compatibility reason.
    This behavior may eventually be dropped for future Gtk major releases.

    Usually, users should not manually perform these initializations but let
    e.g. `Gtk.Application` manage it when needed.
    """
    _options["legacy_autoinit"] = False



# Copyright 2018 Christoph Reiter <reiter.christoph@gmail.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

import warnings

from gi import PyGIDeprecationWarning
from gi.repository import GLib

from ..overrides import override
from ..module import get_introspection_module


GdkPixbuf = get_introspection_module("GdkPixbuf")
__all__ = []


@override
class Pixbuf(GdkPixbuf.Pixbuf):
    @classmethod
    def new_from_data(
        cls,
        data,
        colorspace,
        has_alpha,
        bits_per_sample,
        width,
        height,
        rowstride,
        destroy_fn=None,
        *destroy_fn_data,
    ):
        if destroy_fn is not None:
            w = PyGIDeprecationWarning("destroy_fn argument deprecated")
            warnings.warn(w, stacklevel=2)
        if destroy_fn_data:
            w = PyGIDeprecationWarning("destroy_fn_data argument deprecated")
            warnings.warn(w, stacklevel=2)

        data = GLib.Bytes.new(data)
        return cls.new_from_bytes(
            data, colorspace, has_alpha, bits_per_sample, width, height, rowstride
        )


__all__.append("Pixbuf")



import warnings

from gi import PyGIDeprecationWarning

warnings.warn(
    'gi.pygtkcompat is being deprecated in favor of using "pygtkcompat" directly.',
    PyGIDeprecationWarning,
)

from pygtkcompat import enable


__all__ = ["enable"]



# Copyright 2025 Simon McVittie
# SPDX-License-Identifier: LGPL-2.1-or-later

from ..module import get_introspection_module
from ..overrides import override

from gi.repository import GLib

GioUnix = get_introspection_module("GioUnix")

__all__ = []

if (GLib.MAJOR_VERSION, GLib.MINOR_VERSION) < (2, 86):
    # In older versions of GLib there was some confusion between the
    # platform-specific classes in GioUnix and their equivalents in Gio,
    # resulting in functions like g_desktop_app_info_get_action_name()
    # being assumed to be a global function that happened to take a
    # Gio.DesktopAppInfo first parameter, instead of being a method on a
    # GioUnix.DesktopAppInfo instance. There are not very many classes and
    # methods in GioUnix, so we can wrap them and provide the intended API.

    @override
    class DesktopAppInfo(GioUnix.DesktopAppInfo):
        def get_action_name(self, action_name):
            return GioUnix.DesktopAppInfo.get_action_name(self, action_name)

        def get_boolean(self, key):
            return GioUnix.DesktopAppInfo.get_boolean(self, key)

        def get_categories(self):
            return GioUnix.DesktopAppInfo.get_categories(self)

        def get_filename(self):
            return GioUnix.DesktopAppInfo.get_filename(self)

        def get_generic_name(self):
            return GioUnix.DesktopAppInfo.get_generic_name(self)

        def get_is_hidden(self):
            return GioUnix.DesktopAppInfo.get_is_hidden(self)

        def get_keywords(self):
            return GioUnix.DesktopAppInfo.get_keywords(self)

        def get_locale_string(self, key):
            return GioUnix.DesktopAppInfo.get_locale_string(self, key)

        def get_nodisplay(self):
            return GioUnix.DesktopAppInfo.get_nodisplay(self)

        def get_show_in(self, desktop_env=None):
            return GioUnix.DesktopAppInfo.get_show_in(self, desktop_env)

        def get_startup_wm_class(self):
            return GioUnix.DesktopAppInfo.get_startup_wm_class(self)

        def get_string(self, key):
            return GioUnix.DesktopAppInfo.get_string(self, key)

        def get_string_list(self, key):
            return GioUnix.DesktopAppInfo.get_string_list(self, key)

        def has_key(self, key):
            return GioUnix.DesktopAppInfo.has_key(self, key)

        def launch_action(self, action_name, launch_context=None):
            GioUnix.DesktopAppInfo.launch_action(
                self,
                action_name,
                launch_context,
            )

        def launch_uris_as_manager(
            self,
            uris,
            launch_context,
            spawn_flags,
            user_setup=None,
            user_setup_data=None,
            pid_callback=None,
            pid_callback_data=None,
        ):
            return GioUnix.DesktopAppInfo.launch_uris_as_manager(
                self,
                uris,
                launch_context,
                spawn_flags,
                user_setup,
                user_setup_data,
                pid_callback,
                pid_callback_data,
            )

        def launch_uris_as_manager_with_fds(
            self,
            uris,
            launch_context,
            spawn_flags,
            user_setup,
            user_setup_data,
            pid_callback,
            pid_callback_data,
            stdin_fd,
            stdout_fd,
            stderr_fd,
        ):
            return GioUnix.DesktopAppInfo.launch_uris_as_manager_with_fds(
                self,
                uris,
                launch_context,
                spawn_flags,
                user_setup,
                user_setup_data,
                pid_callback,
                pid_callback_data,
                stdin_fd,
                stdout_fd,
                stderr_fd,
            )

        def list_actions(self):
            return GioUnix.DesktopAppInfo.list_actions(self)

    __all__.append("DesktopAppInfo")

    @override
    class FDMessage(GioUnix.FDMessage):
        def append_fd(self, fd):
            return GioUnix.FDMessage.append_fd(self, fd)

        def get_fd_list(self):
            return GioUnix.FDMessage.get_fd_list(self)

        def steal_fds(self):
            return GioUnix.FDMessage.steal_fds(self)

    __all__.append("FDMessage")

    @override
    class InputStream(GioUnix.InputStream):
        def get_close_fd(self):
            return GioUnix.InputStream.get_close_fd(self)

        def get_fd(self):
            return GioUnix.InputStream.get_fd(self)

        def set_close_fd(self, close_fd):
            GioUnix.InputStream.set_close_fd(self, close_fd)

    __all__.append("InputStream")

    @override
    class MountMonitor(GioUnix.MountMonitor):
        def set_rate_limit(self, limit_msec):
            GioUnix.MountMonitor.set_rate_limit(limit_msec)

    __all__.append("MountMonitor")

    @override
    class OutputStream(GioUnix.OutputStream):
        def get_close_fd(self):
            return GioUnix.OutputStream.get_close_fd(self)

        def get_fd(self):
            return GioUnix.OutputStream.get_fd(self)

        def set_close_fd(self, close_fd):
            GioUnix.OutputStream.set_close_fd(self, close_fd)

    __all__.append("OutputStream")
